--[[
  Plan is quite simple: Implement support for RaidCore Communications.
  This includes:
    - Encounter Communications.
    - Pull/Break Timers.
    - Summon-Accept command.
]]


local ICCommLib = require "ICCommLib"
local ICComm = require "ICComm" -- luacheck: ignore
local GameLib = require "GameLib"
local ApolloTimer = require "ApolloTimer"
local GroupLib = require "GroupLib"
local CSIsLib = require "CSIsLib"
local ChatSystemLib = require "ChatSystemLib"

local Apollo = require "Apollo"
local tAddon = Apollo.GetAddon("MischhBossModules")

local Communicator = {
  bReady = false, -- if this is true, the below will be set.
  tChannel = nil,
  tLUI = nil,
  tJSON = nil,

  --timer/delay for joining the channel.
  nProgressiveRetryDelay = 5, --5, 10, 15, 20, 25, 30 (seconds delay between retry)
  tRetryDelayTimer = nil,
}
tAddon.Communicator = Communicator
tAddon:Link(Communicator) --the ICCommLib does not function otherwise.

local RC_Enum = { --directly copied from RaidCore.
  -- Comm events. Using numbers for these instead to save bandwidth would break
  -- compatibility with older versions of raidcore.
  COMM_VERSION_CHECK_REQUEST                = "VersionCheckRequest", --will probably not happen, ever.
  COMM_VERSION_CHECK_REPLY                  = "VersionCheckReply", --will probably not happen, ever.
  COMM_NEWEST_VERSION                       = "NewestVersion", --will probably not happen, ever.
  COMM_LAUNCH_PULL                          = "LaunchPull", --actually supported
  COMM_LAUNCH_BREAK                         = "LaunchBreak", --actually supported
  COMM_SYNC_SUMMON                          = "SyncSummon", --actually supported
  COMM_ENCOUNTER_IND                        = "Encounter_IND", --actually supported
}
local tCommMethods = {
  [RC_Enum.COMM_VERSION_CHECK_REQUEST]      = "OnVersionCheckRequest", --will probably not happen, ever.
  [RC_Enum.COMM_VERSION_CHECK_REPLY]        = "OnVersionCheckReply", --will probably not happen, ever.
  [RC_Enum.COMM_NEWEST_VERSION]             = "OnNewestVersionRequest", --will probably not happen, ever.
  [RC_Enum.COMM_LAUNCH_PULL]                = "OnPullRequest", --implemented
  [RC_Enum.COMM_LAUNCH_BREAK]               = "OnBreakRequest", --implemented
  [RC_Enum.COMM_SYNC_SUMMON]                = "OnSyncSummonRequest", --implemented
  [RC_Enum.COMM_ENCOUNTER_IND]              = "OnEncounterMessage", --implemented
}

local tCommands = {
  ["RequestSyncSummon"]     = {"summon"},
  ["RequestBreak"]          = {"break"},
  ["RequestPull"]           = {"pull"},
}

local L = tAddon.Localizer{
  ["enUS"] = {
    ["message.summon.csi"] = "Teleport to your group member?",
    ["message.summon.request"] = "%s requested that you accept a summon. Attempting to accept now.",
  },
  ["deDE"] = {
    ["message.summon.csi"] = "Zu deinem Gruppenmitglied teleportieren?",
    ["message.summon.request"] = "%s versucht dich herbeizurufen. Versuche jetzt anzunehmen.",
  },
  ["frFR"] = {
    ["message.summon.csi"]= "Vous téléporter vers le membre du groupe ?",
    ["message.summon.request"] = "%s vous demande d'accepter l'invocation. Attente d'acceptation en cours.",
  },
}

do --Interface for other parts of the addon. (using the communicators core functionality)

  -- ## Encounter Message ##

  function Communicator:SendEncounterMessage(strReason, tData)
    local tMsg = {
      action = RC_Enum.COMM_ENCOUNTER_IND,
      reason = strReason,
      data = tData,
    }
    self:SendMessage(tMsg)
  end

  function Communicator:OnEncounterMessage(tMessage, _) --strSender
    if not self.bReady then return end --else: no self.tLUI
    if not self.tLUI.bIsRunning then return end

    local tEncounter = self.tLUI.tCurrentEncounter
    if not tEncounter then return end
    if not tEncounter.OnCommMessage then return end

    --not sure, if tMessage.sender can be different from strSender, but this is how Raidcore does it.
    tEncounter:OnCommMessage(tMessage.sender, tMessage.reason, tMessage.data)
  end

  -- ## Sync Summon ##

  function Communicator:RequestSyncSummon()
    if not Communicator.IsRaidManagement() then
      self.tLUI:Print("You must be a raid leader or assistant to use this command!")
      return false
    end
    local tMsg = {
      action = RC_Enum.COMM_SYNC_SUMMON,
    }
    self:SendMessage(tMsg)
  end

  function Communicator:OnSyncSummonRequest(tMessage, _) --strSender
    if not Communicator.IsRaidManagement(tMessage.sender) then
      return false
    end

    local CSImsg = CSIsLib.GetActiveCSI()
    if not CSImsg or not CSImsg.strContext then return end

    if CSImsg.strContext == L["message.summon.csi"] then
      self.tLUI:Print(L["message.summon.request"]:format(tMessage.sender))
      if CSIsLib.IsCSIRunning() then
        CSIsLib.CSIProcessInteraction(true)
      end
    end
  end

  -- ## Launch Break ##

  local tBreakAnnouncements = {
    1,2,3,5,10,15,20,30,60,90,120,180,300,600,900,1200
    --1s 2s 3s 5s 10s 15s 20s 30s 1m 1.5m 2m 3m 5m 10m 15m 20m
  }
  local BreakAnnouncer = {
    tTimer = nil,
    tChannel = nil,
    nEndTime = nil,
    nNextUpdate = nil,
  }
  function BreakAnnouncer:Start(nTime)
    self:Abort()

    if not self.tChannel then self:FindChannel() end
    self.tChannel:Send(("Break Timer started: %s"):format(
      BreakAnnouncer.FormatTime(nTime)
    ))

    self:ScheduleUpdate(nTime)
  end

  function BreakAnnouncer:Update()
    if not self.tChannel then self:FindChannel() end

    if self.nNextUpdate > 0 then
      self.tChannel:Send(("Break ends in: %s"):format(
        BreakAnnouncer.FormatTime(self.nNextUpdate)
      ))
      self:ScheduleUpdate(self.nEndTime-GameLib.GetGameTime())
    else
      self.tTimer = nil
      self.nEndTime = nil
      self.nNextUpdate = nil
      self.tChannel:Send("Break ended!")
    end
  end

  function BreakAnnouncer:Abort()
    if self.nEndTime then
      self.tTimer:Stop()
      self.tTimer = nil
      self.nEndTime = nil
      self.nNextUpdate = nil
      self.tChannel:Send("Break aborted!")
    end
  end

  function BreakAnnouncer.FormatTime(nTime)
    local nSecs = math.floor((nTime/1)%60)
    local nMins = math.floor((nTime/60)%60)
    local nHrs = math.floor((nTime/3600))

    local strFormat = ""
    if nHrs > 0 then
      strFormat = ("%d Hours "):format(nHrs)
    end
    if nMins > 0 then
      strFormat = strFormat..("%d Minutes "):format(nMins)
    end
    if nSecs > 0 then
      strFormat = strFormat..("%d Seconds "):format(nSecs)
    end
    return strFormat
  end

  function BreakAnnouncer:ScheduleUpdate(nTimeLeft)
    local nAnnounceLeft = 0
    for i = #tBreakAnnouncements, 1, -1 do
      if tBreakAnnouncements[i] < nTimeLeft-0.3 then
        nAnnounceLeft = tBreakAnnouncements[i]
        break
      end
    end

    local nCurrTime = GameLib.GetGameTime()
    local nTillAnnounce = nTimeLeft-nAnnounceLeft

    self.nEndTime = nCurrTime+nTimeLeft
    self.nNextUpdate = nAnnounceLeft
    self.tTimer = ApolloTimer.Create(nTillAnnounce, false, "Update", self)
  end

  function BreakAnnouncer:FindChannel()
    local tChannels = ChatSystemLib.GetChannels()
    for _, tChannel in pairs(tChannels) do
      if tChannel:GetType() == ChatSystemLib.ChatChannel_Party then
        self.tChannel = tChannel
        return tChannel
      end
    end
  end


  function Communicator:RequestBreak(_, strArgs)
    if not Communicator.IsRaidManagement() then
      self.tLUI:Print("You must be a raid leader or assistant to use this command!")
      return false
    end

    if strArgs then
      strArgs = strArgs:gsub("%s", "") --remove all spaces.
    end

    local nMult; --seconds/minutes.
    local strNum, strMult = strArgs:match("^(%d+)([sm])$")
    if strMult == "m" then
      nMult = 60
      strNum = strNum
    else
      nMult = 1
      strNum = strNum or strArgs
    end

    local nTime = (strNum and tonumber(strNum) or 600)*nMult --default = 10min

    local tMsg = {
      action = RC_Enum.COMM_LAUNCH_BREAK,
      cooldown = nTime,
    }
    self:SendMessage(tMsg)
    self:ProcessMessage(tMsg)
    BreakAnnouncer:Start(nTime)
  end

  function Communicator:OnBreakRequest(tMessage, _) --strSender
    if not Communicator.IsRaidManagement(tMessage.sender) then return end

    if tMessage.sender ~= GameLib.GetPlayerCharacterName() then
      BreakAnnouncer:Abort()
    end

    if tMessage.cooldown and tMessage.cooldown > 0 then
      local nCD = tonumber(tMessage.cooldown)
      self:Timer("break", "Break", nCD, false) --false = no Countdown
      self:Alert("break", ("Break for %s"):format(BreakAnnouncer.FormatTime(nCD)))
    end
  end

  -- ## Pull Timer ##
  local PullAnnouncer = {
    tChannel = nil,
    tTimer = nil,
    nEndTime = nil,
  }
  function PullAnnouncer:Start(nTime)
    self:Abort()
    if not self.tChannel then self:FindChannel() end
    self.nEndTime = GameLib.GetGameTime() + nTime
    self:Update()
  end

  function PullAnnouncer:Update()
    if not self.tChannel then self:FindChannel() end

    local nCurr = GameLib.GetGameTime()
    local nLeft = self.nEndTime-nCurr

    --round to closest number:
    local nAnnounce = math.floor(nLeft+0.5)
    local nNextUpdate = nAnnounce-1 -- next number to be announced.
    local nTillAnnounce = nLeft - nNextUpdate -- ~1sec

    if nAnnounce > 0 then
      self.tChannel:Send(("Pulling in... %d"):format(nAnnounce))
      self.tTimer = ApolloTimer.Create(nTillAnnounce, false, "Update", self)
    else --PULLING!
      self.tChannel:Send("Pulling!")
      self.tTimer = nil
      self.nEndTime = nil
    end
  end

  function PullAnnouncer:Abort()
    if self.nEndTime then
      self.tTimer:Stop()
      self.tTimer = nil
      self.nEndTime = nil
      self.tChannel:Send("Pull aborted!")
    end
  end

  function PullAnnouncer:FindChannel()
    local tChannels = ChatSystemLib.GetChannels()
    for _, tChannel in pairs(tChannels) do
      if tChannel:GetType() == ChatSystemLib.ChatChannel_Party then
        self.tChannel = tChannel
        return tChannel
      end
    end
  end


  function Communicator:RequestPull(_, strArgs)
    if not Communicator.IsRaidManagement() then
      self.tLUI:Print("You must be a raid leader or assistant to use this command!")
      return false
    end

    if strArgs then
      strArgs = strArgs:gsub("%s", "") --remove all spaces.
    end

    local nTime = math.ceil(tonumber(strArgs) or 5)
    local tMsg = {
      action = RC_Enum.COMM_LAUNCH_PULL,
      cooldown = nTime,
    }
    self:SendMessage(tMsg)
    self:ProcessMessage(tMsg)
    PullAnnouncer:Start(nTime)
  end

  function Communicator:OnPullRequest(tMessage, _) --strSender
    if not Communicator.IsRaidManagement(tMessage.sender) then return end

    if tMessage.sender ~= GameLib.GetPlayerCharacterName() then
      PullAnnouncer:Abort()
    end

    if tMessage.cooldown and tMessage.cooldown > 0 then
      local nCD = tonumber(tMessage.cooldown)
      self:Timer("pull", "Pull", nCD, true) --false = no Countdown
      self:FakeCast("Pulling", nCD)
    end
  end
end


do --Core Functionality - provides :SendMessage(oMsg, oPlayer) and calls functions according to the above.
  function Communicator:OnDocLoaded()
    if tAddon.Injector:HasInjected() then
      for strHandler, arSlashCommands in pairs(tCommands) do
        for _, strCommand in ipairs(arSlashCommands) do
          Apollo.RegisterSlashCommand(strCommand, strHandler, self)
        end
      end

      self:JoinChannelTry()
    end
  end
  tAddon:OnceDocLoaded("OnDocLoaded", Communicator)

  function Communicator:JoinChannelTry()
    local eChannelType = ICCommLib.CodeEnumICCommChannelType.Group
    local sChannelName = "RaidCore"

    --request to join the channel.
    self.tChannel = ICCommLib.JoinChannel(sChannelName, eChannelType)
    --start the timer for retry.
    self.tRetryDelayTimer = ApolloTimer.Create(self.nProgressiveRetryDelay, false, "JoinChannelTry", self)
    self.nProgressiveRetryDelay = math.min(self.nProgressiveRetryDelay + 5, 30)

    if self.tChannel then
      if self.tChannel:IsReady() then
        self:OnJoinChannelResult(self.tChannel, ICCommLib.CodeEnumICCommJoinResult.Join)
      else
        self.tChannel:SetJoinResultFunction("OnJoinChannelResult", self)
      end
    end
  end

  function Communicator:OnJoinChannelResult(tChannel, eResult)
    if eResult == ICCommLib.CodeEnumICCommJoinResult.Join then
      if self.tRetryDelayTimer then
        self.tRetryDelayTimer:Stop()
        self.tRetryDelayTimer = nil
      end

      tChannel:SetReceivedMessageFunction("OnRecievedMessage", self)
      tChannel:SetSendMessageResultFunction("OnSendMessageResult", self)

      self.bReady = true
      self.tChannel = tChannel
      self.tLUI = tAddon.Injector:GetLUI()
      self.tJSON = Apollo.GetPackage("Lib:dkJSON-2.5").tPackage
    end
    --else: self.tRetryDelayTimer will trigger sometime later.
  end

  function Communicator:OnRecievedMessage(--[[tChannel]] _, strMessage, strSender)
    if not self.bReady then return end --dont have self.tJSON else

    local tMessage = self.tJSON.decode(strMessage)
    self:ProcessMessage(tMessage, strSender)
  end

  function Communicator:OnSendMessageResult(...) --luacheck:ignore
    --Raidcore does some logging here.
  end


  function Communicator:ProcessMessage(tMessage, strSender)
    if type(tMessage) ~= "table" or type(tMessage.action) ~= "string" then
      -- Silent error.
      return
    end

    local strMethod = tCommMethods[tMessage.action]
    if not strMethod then return end
    local fAction = self[strMethod]
    if not fAction then return end

    fAction(self, tMessage, strSender)
  end

  --[[
    oMessage may be:
      - string: to be sent as such.
      - table: to be sent through JSON and then sent.

    oPlayer may be:
      - nil: broadcast the message (only groupwide)
      - string: direct name of unit to send it to
      - userdata: the unit to be sent to
      - number: unitID to send it to
  ]]
  function Communicator:SendMessage(oMessage, oPlayer)
    if not self.bReady then return end --dont have self.JSON, self.tChannel

    local typMessage, typPlayer = type(oMessage), type(oPlayer)
    assert(typMessage == "string" or typMessage == "table")
    assert(typPlayer == "nil" or typPlayer == "string"
      or typPlayer == "userdata" or typPlayer == "number")


    -- Conversion to strMessage, strPlayer (strPlayer may be nil.)
    local strMessage, strPlayer;
    if typMessage == "string" then
      strMessage = oMessage
    elseif typMessage == "table" then
      oMessage.sender = GameLib.GetPlayerCharacterName()
      strMessage = self.tJSON.encode(oMessage)
    end

    if typPlayer == "nil" then --just broadcast
      strPlayer = nil
    elseif typPlayer == "string" then
      local tPlayer = GameLib.GetPlayerUnitByName(oPlayer)
      if not tPlayer or not tPlayer:IsInYourGroup() then return end --send aborted: invalid unit.

      strPlayer = oPlayer
    elseif typPlayer == "userdata" then
      if not oPlayer:IsInYourGroup() then return end --send aborted: invalid unit.

      strPlayer = oPlayer:GetName()
    elseif typPlayer == "number" then
      local tPlayer = GameLib.GetUnitById(oPlayer)

      if not tPlayer then return end --send aborted: unknown unitID
      if not tPlayer:IsInYourGroup() then return end --send aborted: invalid unit.

      strPlayer = tPlayer:GetName()
    end

    if strPlayer then
      self.tChannel:SendPrivateMessage(strPlayer, strMessage)
    else
      self.tChannel:SendMessage(strMessage)
    end
  end
end

do --Misc Helper Functions
  --[[
    oPlayer may be:
      nil: the player himself
      string: a player with this name
      number: the unit with this id
      userdata: the specific unit
  ]]
  function Communicator.IsRaidManagement(oPlayer)
    if not GroupLib.InGroup() then return false end
    local strPlayer;

    local typPlayer = type(oPlayer)
    if typPlayer == "nil" then
      strPlayer = GameLib.GetPlayerCharacterName()
    elseif typPlayer == "string" then
      strPlayer = oPlayer
    elseif typPlayer == "number" then
      local tPlayer = GameLib.GetUnitById(oPlayer)
      if not tPlayer or not tPlayer:IsInYourGroup() then return false end
      strPlayer = tPlayer:GetName()
    elseif typPlayer == "userdata" then
      if not oPlayer:IsInYourGroup() then return false end
      strPlayer = oPlayer:GetName()
    end

    for nIdx=1, GroupLib.GetMemberCount(), 1 do
      local tGroupMember = GroupLib.GetGroupMember(nIdx)
      if tGroupMember and tGroupMember.strCharacterName == strPlayer then
        if tGroupMember.bIsLeader or tGroupMember.bRaidAssistant then
          return true
        else
          return false
        end
      end
    end
    return false --just in case.
  end

  local LUI_Updater = {}
  LUI_Updater.__index = LUI_Updater
  function LUI_Updater.new(tLUI, arUpdateMethods)
    local tUpdater = setmetatable({
      tTimer = nil,
      tArg = nil, --set in :Engage
      tLUI = tLUI,

      arUpdates = arUpdateMethods,
    }, LUI_Updater)

    return tUpdater
  end

  function LUI_Updater:GetRemoveFunction(strFunc, tHandler, ...)
    local fCallback;
    if strFunc then
      fCallback = tAddon.Utils.CreateCallback(0, strFunc, tHandler, ...)
    end
    return function()
      self.tTimer:Stop()
      self.tTimer = nil
      if fCallback then fCallback() end
    end
  end

  function LUI_Updater:Update()
    for _, strMethod in ipairs(self.arUpdates) do
      self.tLUI[strMethod](self.tLUI, self.tArg)
    end
  end

  function LUI_Updater:Engage(tArg)
    self.tArg = tArg
    self.tTimer = ApolloTimer.Create(0.1, true, "Update", self)
  end


  function Communicator:Timer(key, strText, nTime, bCountdown, strFunc, tHandler, ...)
    if not self.bReady then return end

    local tUpdater = LUI_Updater.new(self.tLUI, {"UpdateTimer", "SortTimer"})
    local fRemove = tUpdater:GetRemoveFunction(strFunc, tHandler, ...)

    self.tLUI:RemoveTimer(key)
    self.tLUI:AddTimer(key, strText, nTime, {enable=true, sound=bCountdown}, fRemove, "break")

    local tArg = self.tLUI.runtime.timer[key]
    if tArg then
      tUpdater:Engage(tArg)
    end
  end

  function Communicator:Alert(key, strText, strFunc, tHandler, ...)
    if not self.bReady then return end

    local tUpdater = LUI_Updater.new(self.tLUI, {"UpdateAlert"})
    local fRemove = tUpdater:GetRemoveFunction(strFunc, tHandler, ...)

    self.tLUI:HideAlert(key)
    self.tLUI:ShowAlert(key, strText, {enable=true}, fRemove, "break")

    local tArg = self.tLUI.runtime.alerts[key]
    if tArg then
      tUpdater:Engage(tArg)
    end
  end

  function Communicator:FakeCast(strText, nDuration, strFunc, tHandler, ...)
    if not self.bReady then return end

    local tUpdater = LUI_Updater.new(self.tLUI, {"UpdateCast"})
    local fRemove = tUpdater:GetRemoveFunction(strFunc, tHandler, ...)

    local tCast = setmetatable({
      sName = "MOO", --this makes the bar count down instead of up.
      nDuration = nDuration,
      nStart = GameLib.GetGameTime(),
    }, {__index = function(t, k)
      if k == "nElapsed" then
        return GameLib.GetGameTime() - t.nStart
      end
    end})

    if self.tLUI.runtime and self.tLUI.runtime.cast then
      self.tLUI:HideCast()
    end
    self.tLUI:ShowCast(tCast,strText,{enable=true}, fRemove, "break")
    tUpdater:Engage(nil)
  end
end