local Apollo = require "Apollo"

-----------------------------------------------------------------------------------------------
-- Module Definition
-----------------------------------------------------------------------------------------------
local tAddon = Apollo.GetAddon("MischhBossModules")
--this module only does things, if the addon 'Rover' is supplied.

local Debug = {
  tCache = {}, --caching for later output to rover
  bPcall = false,--whether or not we are currently in a pcall()
  tRegistered = {}, --all registered calls, which have not been passed on.
  bEnabled = false, --whether or not tRegistered was passed on
}
tAddon.Debug = Debug

-----------------------------------------------------------------------------------------------
-- Debug Module
-----------------------------------------------------------------------------------------------
--Arguments are either a table, or a "path" to a table, originating in tAddon
--example: Debug:Add("Identifier", "A", "B") would be the same as Debug:Add("Identifier", tAddon.A.B)
function Debug:Add(Id, ...)
  if type(Id) ~= "string" then return end
  local tTarget = self.GetTable(...)

  if not tTarget then error("Target table not found") end

  local tOverwritten = {}

  for tTable in self.Metatables(tTarget) do
    for i,v in pairs(tTable) do
      if type(i) == "string" and i:sub(1,2) ~= "__" and type(v) == "function" and not tOverwritten[i] then
        tTarget[i] = self:CreateReplacement(Id..":"..i, v)
      end
      tOverwritten[i] = true
    end
  end
end

function Debug:Enable()
  if self.bEnabled then return end
  for _, tArgs in ipairs(self.tRegistered) do
    self:Add(unpack(tArgs))
  end
  self.bEnabled = true
end

--does Debug:Add(...) once Debug:Enable() was called
function Debug:Register(...)
  if self.bEnabled then
    self:Add(...)
  else
    table.insert(self.tRegistered, {...})
  end
end

-------------------------------------------------------------------------------------------------
-- Private Functions
-------------------------------------------------------------------------------------------------
function Debug.Rover(...) -- this will be replaced by _G.SendVarToRover, after Addon gets loaded.
  local args = {...}
  table.insert(Debug.tCache, args)
end

function Debug:OnLoad()
  if _G.SendVarToRover then
    self.Rover = _G.SendVarToRover

    --push the cache to Rover.
    for _, tArgs in ipairs(self.tCache) do
      self.Rover(unpack(tArgs))
    end

    --clear the cache: keep an empty cache table to be safe.
    self.tCache = {}
  else
    --Rover doesnt exist -> assume this addon should not be debugged.
    self.Rover = function() end
    self.tCache = {}
  end
end
tAddon:OnceLoaded("OnLoad", Debug)


function Debug:CreateReplacement(strName, fFunction)
  return function(...)
    local tArgs = {...}
    local tReturns;

    local strError, strTrace;
    local tData = {}
    self.Rover(strName, tData, 0)

    if not self.bPcall then
      self.bPcall = true
      xpcall(function() tReturns = {fFunction(unpack(tArgs))} end,
        function(err) strError, strTrace = err, debug.traceback(err, 2) end)
      self.bPcall = false
    else
      tReturns = {fFunction(...)}
    end

    if tArgs and #tArgs > 0 then
      tData.arg = tArgs
    end

    if tReturns and #tReturns > 0 then
      tData.ret = tReturns
    end

    if strError then
      tData.err = strError
      tData.trace = strTrace
    end

    if strError then
      self.Rover("<ERR>"..strName, tData, 0)
      _G.Print("<ERROR>"..strError)
    end

    return unpack(tReturns)
  end
end

function Debug.GetTable(...)
  if type((...)) == "table" then
    return (...)
  end

  local tTarget = tAddon

  for _, idx in ipairs({...}) do
    if type(tTarget[idx]) == "table" then
      tTarget = tTarget[idx]
    else
      error(("Invalid Path: expected table, got: %s"):format(type(tTarget[idx])))
    end
  end

  return tTarget
end

function Debug.Metatables(tTbl)
  local returned = {}
  local tLast;
  return function()
    if not tLast then
      tLast = tTbl
      returned[tLast] = true
      return tLast
    else
      local tNext = getmetatable(tLast)
      tNext = type(tNext) == "table" and tNext.__index
      if type(tNext) ~= "table" or returned[tNext] then
        return nil
      else
        tLast = tNext
        returned[tLast] = true
        return tLast
      end
    end
  end
end
