local Apollo = require "Apollo"
local GameLib = require "GameLib"
local ApolloTimer = require "ApolloTimer"
local Vector3 = require "Vector3"
local GroupLib = require "GroupLib"

local tAddon = Apollo.GetAddon("MischhBossModules")

local Encounter = {}
tAddon.Encounter = Encounter

local function fCallbackHandler(strEvent)
	return function(self, tData)
		self:FireCallbacks(strEvent, {
			tData,
			tData = tData,
		}, tData)
	end
end

function Encounter:FormatLocalize(oFormat)
	if type(oFormat) == "string" then
		return self.L[oFormat]
	elseif type(oFormat) == "table" then
		local strFormat = self.L[tostring(oFormat[1])]
		return strFormat:format(select(2, unpack(oFormat)))
	else
		return oFormat
	end
end

-- Key, > tUnit <, KeyExt, tOverrides
function Encounter:Unit(strKey, tUnit, strKeyExt, tOverrides)
	if not self.config.units or not self.config.units[strKey] then return end
	--translate to:
	--self.core:AddUnit(nId, sName, tUnit, tConfig, sMark)
	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.units[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	if type(tUnit) == "number" then
		tUnit = GameLib.GetUnitById(tUnit)
	end

	local strMark = self:FormatLocalize(tOverrides and tOverrides.strMark) or nil
	local strName = self:FormatLocalize(tOverrides and tOverrides.strName) or tUnit:GetName()
	local nId = tUnit:GetId()

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end
	--core:AddUnit(nId) REQUIRES you to enter the units Id! - thats why we manually track the conversion
	--		between strKey and nId. We do this for :GetUnit(strKey) and trying to stay with a uniform style.

	if self.tTrackedUnits[strKey] and self.tTrackedUnits[strKey] ~= nId then
		self.core:RemoveUnit(self.tTrackedUnits[strKey])
	end

	self.tTrackedUnits[strKey] = nId
	return self.core:AddUnit(nId, strName, tUnit, tConfig, strMark)
end

function Encounter:RemoveUnit(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	local nId = self.tTrackedUnits[strKey] or strKey

	if self.core:RemoveUnit(nId) then
		self.tTrackedUnits[strKey] = nil
		return true
	else
		return false
	end
end

function Encounter:GetUnit(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	local nId = self.tTrackedUnits[strKey] or strKey

	if self.core and self.core.runtime and self.core.runtime.units then
		return (self.core.runtime.units[nId] or {}).tUnit
	end
end

function Encounter:HasUnit(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	local nId = self.tTrackedUnits[strKey] or strKey

	if self.core and self.core.runtime and self.core.runtime.units then
		local tUnit = (self.core.runtime.units[nId] or {}).tUnit
		return tUnit and true or false
	else
		return false
	end
end

function Encounter:GetUnitId(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	local nId = self.tTrackedUnits[strKey]

	return nId
end

function Encounter:GetHealth(strKey, strKeyExt, tOverrides)
	--first: try :GetUnit()
	local tUnit = self:GetUnit(strKey, strKeyExt, tOverrides)

	if not tUnit then
		if type(strKey) == "userdata" then --assume its the unit itself.
			tUnit = strKey
		elseif type(strKey) == "number" then
			tUnit = GameLib.GetUnitById(strKey)
		end
	end

	if tUnit then
		if not tUnit:IsValid() then return 0 end --assume dead, if invalid.
		local nHealth = tUnit:GetHealth() or 0
		local nMaxHealth = math.max(tUnit:GetMaxHealth() or 1, 1)

		return nHealth/nMaxHealth*100
	end

	return 0
end

function Encounter:TrackUnit(...) --add a unit with disabled 'enable'-flag
	if type((...)) == "userdata" then
		local tUnit = ...
		return self.core:AddUnit(tUnit:GetId(), tUnit:GetName(), tUnit, {enable=false}, nil)
	else
		local strKey, tUnit, strKeyExt, tOverrides = ...
		tOverrides = tOverrides or {}
		tOverrides.enable = false

		return self:Unit(strKey, tUnit, strKeyExt, tOverrides)
	end
end

function Encounter:Timer(strKey, nDuration, strText, strKeyExt, tOverrides)
	if not self.config.timers or not self.config.timers[strKey] then return end
	--translate to:
	--self.core:AddTimer(strName, strText, nDuration, tConfig, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.timers[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	strText = self:FormatLocalize(strText or tOverrides and tOverrides.strText or tConfig.label)
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end


	self.core:AddTimer(strKey, strText, nDuration, tConfig, fHandler, tData)
end

function Encounter:RemoveTimer(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:RemoveTimer(sName)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:RemoveTimer(strKey)
end

function Encounter:Cast(strKey, tCast, strKeyExt, tOverrides) --luacheck: no unused args
	if not self.config.casts or not self.config.casts[strKey] then return end
	--translate to:
	--self.core:ShowCast(tCast, sName, tConfig, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.casts[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	if type(tCast) == "number" then
		local nDuration = tCast
		tCast = setmetatable({
			sName = self.L[tConfig.label],
			nDuration = nDuration,
			nStart = GameLib.GetGameTime(),
		}, {__index = function(t, k)
			if k == "nElapsed" then
				return GameLib.GetGameTime() - t.nStart
			end
		end})
	end

	local strName = self:FormatLocalize(tOverrides and tOverrides.strText)
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	self.core:ShowCast(tCast, strName, tConfig, fHandler, tData)
end

function Encounter:RemoveCast(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:HideCast()
	self.core:HideCast() --wow. that was easy ^^
end

function Encounter:Aura(strKey, nDuration, strKeyExt, tOverrides)
	if not self.config.auras or not self.config.auras[strKey] then return end
	--translate to:
	--self.core:ShowAura(sName, tConfig, nDuration, sText, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.auras[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	local strText = self:FormatLocalize(tOverrides and tOverrides.strText or tConfig.label)
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:ShowAura(strKey, tConfig, nDuration, strText, fHandler, tData)
end
function Encounter:RemoveAura(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:HideAura(sName)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:HideAura(strKey)
end
function Encounter:Alert(strKey, strText, strKeyExt, tOverrides)
	if not self.config.alerts or not self.config.alerts[strKey] then return end
	--translate to:
	--self.core:ShowAlert(sName, sText, tConfig, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.alerts[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	strText = self:FormatLocalize(strText or tOverrides and tOverrides.strText or tConfig.label)
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end


	self.core:ShowAlert(strKey, strText, tConfig, fHandler, tData)
end
function Encounter:RemoveAlert(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:HideAlert(sName)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:HideAlert(strKey)
end
function Encounter:Sound(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	if not self.config.sounds or not self.config.sounds[strKey] then return end
	--translate to:
	--self.core:PlaySound(tConfig)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.sounds[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	self.core:PlaySound(tConfig)
end
function Encounter:Text(strKey, tOrigin, strText, strKeyExt, tOverrides)
	if not self.config.texts or not self.config.texts[strKey] then return end
	--translate to:
	--self.core:DrawText(Key, Origin, tConfig, sText, bTop, nOffset, nDuration, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.texts[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	strText = self:FormatLocalize(strText or tOverrides and tOverrides.strText or tConfig.label)
	local bTop = tOverrides and tOverrides.bTop or nil
	local nOffset = tOverrides and tOverrides.nOffset or nil
	local nDuration = tOverrides and tOverrides.nDuration or nil
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local fUpdate = tOverrides and tOverrides.fUpdate or nil
	local tData = tOverrides and tOverrides.tData or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end
	strKey = "Text"..strKey

	local tOrigDraw; --for the Update function.
	local tHandler;
	if fUpdate then
		tHandler = {OnFrame = function()
			if not self.core then return end
			if not self.core.tDraws then return end
			if not self.core.tDraws[strKey] then return end
			if not self.core.tDraws[strKey].wnd then return end
			if self.core.tDraws[strKey]~=tOrigDraw then
				return Apollo.RemoveEventHandler("NextFrame", tHandler)
			end

			local tDraw = self.core.tDraws[strKey]

			local strUpdateText = fUpdate(self, tDraw, tData)

			if strUpdateText ~= nil then
				tDraw.wnd:SetText(strUpdateText)
			end
		end} Apollo.RegisterEventHandler("NextFrame", "OnFrame", tHandler)

		local fOrigHandler = fHandler
		fHandler = function(...)
			Apollo.RemoveEventHandler("NextFrame", tHandler)
			if type(fOrigHandler) == "function" then
				return fOrigHandler(...)
			end
		end
	end

	self.core:DrawText(strKey, tOrigin, tConfig, strText, bTop, nOffset, nDuration, fHandler, tData)

	if tHandler then
		tOrigDraw = self.core.tDraws[strKey]
		tHandler:OnFrame()
	end
end
function Encounter:RemoveText(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:RemoveText(Key)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:RemoveText("Text"..strKey)
end

function Encounter:TimerText(strKey, tOrigin, nDuration, strKeyExt, tOverrides)
	if nDuration and nDuration>0 then
		tOverrides = tOverrides or {}
		tOverrides.timer = true
		tOverrides.nDuration = nDuration

		self:Text(strKey, tOrigin, "77,8", strKeyExt, tOverrides)
	end
end

function Encounter:RemoveTimerText(...)
	return self:RemoveText(...)
end

function Encounter:HealthText(strKey, tOrigin, nHealthTarget, strKeyExt, tOverrides)
	if type(tOrigin) == "number" then tOrigin = GameLib.GetUnitById(tOrigin) end
	if type(tOrigin) ~= "userdata" then return end

	nHealthTarget = nHealthTarget or 0

	tOverrides = tOverrides or {}

	local fOrigUpdate = tOverrides.fUpdate
	tOverrides.fUpdate = function(this, tDraw, ...)
		local tUnit = tDraw.tOrigin
		if tUnit:IsValid() then
			local nHealth = tUnit:GetHealth()
			local nMaxHealth = tUnit:GetMaxHealth()
			if nMaxHealth < 1 then nMaxHealth = 1 end
			local nPercent = (100*nHealth/nMaxHealth)-nHealthTarget

			local strText;
			if nPercent <= 0 then
				return self:RemoveText(strKey, strKeyExt)
			elseif nPercent < 10 then
				strText = Apollo.FormatNumber(nPercent, 1, true).."%"
			else
				strText = Apollo.FormatNumber(nPercent, 0, true).."%"
			end

			tDraw.wnd:SetText(strText)
			if type(fOrigUpdate) == "function" then
				return fOrigUpdate(this, tDraw, ...)
			end
		end
	end

	return self:Text(strKey, tOrigin, "7,7%", strKeyExt, tOverrides)
end

function Encounter:RemoveHealthText(...)
	self:RemoveText(...)
end

function Encounter:Icon(strKey, tOrigin, strKeyExt, tOverrides)
	if not self.config.icons or not self.config.icons[strKey] then return end
	--translate to:
	--self.core:DrawIcon(Key, Origin, tConfig, bTop, nOffset, nDuration, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.icons[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	local bTop = tOverrides and tOverrides.bTop  or nil
	local nOffset = tOverrides and tOverrides.nOffset  or nil
	local nDuration = tOverrides and tOverrides.nDuration or nil
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end
	strKey = "Icon"..strKey

	self.core:DrawIcon(strKey, tOrigin, tConfig, bTop, nOffset, nDuration, fHandler, tData)
end
function Encounter:RemoveIcon(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:RemoveIcon(Key)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:RemoveIcon("Icon"..strKey)
end
function Encounter:Polygon(strKey, tOrigin, nRadius, nSide, strKeyExt, tOverrides)
	if not self.config.lines or not self.config.lines[strKey] then return end
	--translate to:
	--self.core:DrawPolygon(Key, Origin, tConfig, nRadius, nRotation, nSide, nDuration, tVectorOffsets, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.lines[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	local nRotation = tOverrides and tOverrides.nRotation or nil
	local tVectorOffsets = tOverrides and tOverrides.tVectorOffsets or nil
	local nDuration = tOverrides and tOverrides.nDuration or nil
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end
	strKey = "Polygon"..strKey

	self.core:DrawPolygon(strKey, tOrigin, tConfig, nRadius, nRotation, nSide, nDuration, tVectorOffsets, fHandler, tData)
end
function Encounter:RemovePolygon(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:RemovePolygon(Key)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:RemovePolygon("Polygon"..strKey)
end
function Encounter:Line(strKey, tOrigin, nLength, strKeyExt, tOverrides)
	if not self.config.lines or not self.config.lines[strKey] then return end
	--translate to:
	--self.core:DrawLine(Key, Origin, tConfig, nLength, nRotation, nOffset, tVectorOffsets, tVectorFacing,
	--nDuration, nNumberOfDot, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.lines[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	local nRotation = tOverrides and tOverrides.nRotation or nil
	local tVectorOffsets = tOverrides and tOverrides.tVectorOffsets or nil
	local nDuration = tOverrides and tOverrides.nDuration or nil
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil
	local nOffset = tOverrides and tOverrides.nOffset or nil
	local tVectorFacing = tOverrides and tOverrides.tVectorFacing or nil
	local nNumberOfDot = tOverrides and tOverrides.nNumberOfDot or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end
	strKey = "Line"..strKey

	self.core:DrawLine(strKey, tOrigin, tConfig, nLength, nRotation, nOffset, tVectorOffsets, tVectorFacing,
				nDuration, nNumberOfDot, fHandler, tData)
end
function Encounter:RemoveLine(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:RemoveLine(Key)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:RemoveLine("Line"..strKey)
end
function Encounter:LineBetween(strKey, tFromOrigin, tToOrigin, strKeyExt, tOverrides)
	if not self.config.lines or not self.config.lines[strKey] then return end
	--translate to:
	--self.core:DrawLineBetween(Key, FromOrigin, ToOrigin, tConfig, nDuration, nNumberOfDot, fHandler, tData)

	local tConfig = tOverrides or {}
	for i,v in pairs(self.config.lines[strKey] or {}) do
		if tConfig[i] == nil then
			tConfig[i] = v
		end
	end

	local nDuration = tOverrides and tOverrides.nDuration or nil
	local fHandler = tOverrides and tOverrides.fHandler or nil
	local tData = tOverrides and tOverrides.tData or nil
	local nNumberOfDot = tOverrides and tOverrides.nNumberOfDot or nil
	local nOffset = tOverrides and tOverrides.nOffset or nil
	local nLength = tOverrides and tOverrides.nLength or nil

	if type(fHandler) == "string" then
		fHandler = fCallbackHandler(fHandler)
	end

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end
	strKey = "LineBetween"..strKey


	-- ### Overly complicated implementations for nOffset and nLength. ###
	local bAdjustFromType = false
	if nOffset then
		local bFromFixed = type(tFromOrigin) == "table" or Vector3.Is(tFromOrigin)
		local bToFixed = type(tToOrigin) == "table" or Vector3.Is(tToOrigin)

		if bFromFixed and bToFixed then --this is the only case, in which we could calculate a static offset.
			local v3From = Vector3.New(tFromOrigin)
			local v3Normal = (Vector3.New(tToOrigin) - v3From):Normal()

			tFromOrigin = v3From + v3Normal*nOffset
		else --we need to create a unit.
			local tFrom = tFromOrigin --could be 'fixed' a 'unitId' or a 'unit'
			if bFromFixed then
				tFrom = Vector3.New(tFromOrigin)
			elseif type(tFromOrigin) == "number" then
				tFrom = GameLib.GetUnitById(tFromOrigin)
			elseif not tFrom then
				tFrom = GameLib.GetPlayerUnit()
			end

			local	tTo = tToOrigin
			if bToFixed then
				tTo = Vector3.New(tToOrigin)
			elseif type(tToOrigin) == "number" then
				tTo = GameLib.GetUnitById(tToOrigin)
			elseif not tTo then
				tTo = GameLib.GetPlayerUnit()
			end

			local tBaseUnit = not bFromFixed and tFrom or tTo --one of them has to be a unit. (prefer From)

			tFromOrigin = setmetatable({
				GetPosition = function()
					local v3From = bFromFixed and tFrom or tFrom:GetPosition()
					if not bFromFixed then
						if v3From then v3From = Vector3.New(v3From)
						else return end
					end

					local v3To = bToFixed and tTo or tTo:GetPosition()
					if not bToFixed then
						if v3To then v3To = Vector3.New(v3To)
						else return v3From end
					end

					local v3Normal = (v3To - v3From):Normal() --vector3 expected, got table?
					return v3From + v3Normal*nOffset
				end
			}, {__index = function(tFake, strMethod)
				tFake[strMethod] = function(_, ...)
					return tBaseUnit[strMethod](tBaseUnit, ...)
				end
				return tFake[strMethod]
			end})
			bAdjustFromType = true
		end
	end

	local bAdjustToType = false
	if nLength then --we already handled nOffset, so 'tFromOrigin' already has the offset accounted into it.
		local bFromFixed = (type(tFromOrigin) == "table" or Vector3.Is(tFromOrigin)) and not bAdjustFromType or false
		local bToFixed = type(tToOrigin) == "table" or Vector3.Is(tToOrigin)

		if bFromFixed and bToFixed then --this is the only case, in which we could calculate a static offset.
			local v3From = Vector3.New(tFromOrigin)
			local v3Normal = (Vector3.New(tToOrigin) - v3From):Normal()

			tToOrigin = v3From + v3Normal*nLength
		else --we need to create a unit.
			local tFrom = tFromOrigin --could be 'fixed' a 'unitId' or a 'unit'
			if bFromFixed then
				tFrom = Vector3.New(tFromOrigin)
			elseif type(tFromOrigin) == "number" then
				tFrom = GameLib.GetUnitById(tFromOrigin)
			elseif not tFrom then
				tFrom = GameLib.GetPlayerUnit()
			end

			local	tTo = tToOrigin
			if bToFixed then
				tTo = Vector3.New(tToOrigin)
			elseif type(tToOrigin) == "number" then
				tTo = GameLib.GetUnitById(tToOrigin)
			elseif not tTo then
				tTo = GameLib.GetPlayerUnit()
			end

			local tBaseUnit = not bToFixed and tTo or tFrom --one of them has to be a unit. (prefer To)

			tToOrigin = setmetatable({
				GetPosition = function()
					local v3To = bToFixed and tTo or tTo:GetPosition()
					if not bToFixed then
						if v3To then v3To = Vector3.New(v3To)
						else return end
					end

					local v3From = bFromFixed and tFrom or tFrom:GetPosition()
					if not bFromFixed then
						if v3From then v3From = Vector3.New(v3From)
						else return v3To end
					end

					local v3Normal = (v3To - v3From):Normal()
					return v3From + v3Normal*nLength
				end
			}, {__index = function(tFake, strMethod)
				tFake[strMethod] = function(_, ...)
					return tBaseUnit[strMethod](tBaseUnit, ...)
				end
				return tFake[strMethod]
			end})
			bAdjustToType = true
		end
	end


	local fOrigType = _G.type
	if bAdjustFromType or bAdjustToType then
		if bAdjustFromType and bAdjustToType then
			_G.type = function(...)
				if (...) == tFromOrigin or (...) == tToOrigin then
					return "userdata"
				else
					return fOrigType(...)
				end
			end
		elseif bAdjustFromType then
			_G.type = function(...)
				if (...) == tFromOrigin then
					return "userdata"
				else
					return fOrigType(...)
				end
			end
		else --bAdjustToType
			_G.type = function(...)
				if (...) == tToOrigin then
					return "userdata"
				else
					return fOrigType(...)
				end
			end
		end
	end

	self.core:DrawLineBetween(strKey, tFromOrigin, tToOrigin, tConfig, nDuration, nNumberOfDot, fHandler, tData)
	_G.type = fOrigType
end


function Encounter:RemoveLineBetween(strKey, strKeyExt, tOverrides) --luacheck: no unused args
	--translate to:
	--self.core:RemoveLineBetween(Key)

	if strKeyExt then
		strKey = strKey..tostring(strKeyExt)
	end

	self.core:RemoveLineBetween("LineBetween"..strKey)
end

function Encounter:SendCommMessage(strReason, tData) --luacheck: no unused args (self)
	tAddon.Communicator:SendEncounterMessage(strReason, tData)
end

function Encounter:Print(strMessage)
	strMessage = self:FormatLocalize(strMessage)
	if strMessage then
		return self.core:Print(strMessage)
	end
end

do --DelayedEvent
	local Delay = {}
	Delay.__index = Delay

	function Delay:Trigger()
		self.tEncounter.tRunningTimers[self.nId] = nil

		for _, oDelay in pairs(self.tDelays) do
			if type(oDelay) == "table" then
				local oMethod = oDelay[1]
				if type(oMethod) == "function" then
					oMethod(select(2, unpack(oDelay)))
				elseif type(oMethod) == "string" then
					self.tEncounter[oMethod](self.tEncounter, select(2, unpack(oDelay)))
				end
			elseif type(oDelay) == "function" then
				oDelay(self.tEncounter)
			elseif type(oDelay) == "string" then
				self.tEncounter[oDelay](self.tEncounter)
			end
		end
	end

	function Encounter:Delay(strKey, nDelay, ...)
		if type(strKey) == "number" then
			return self:Delay(nil, strKey, nDelay, ...)
		end

		local nId = strKey or #self.tRunningTimers+1

		self:RemoveDelay(nId)

		local tHandler = setmetatable({
			tEncounter = self,
			tDelays = {...},
			nId = nId,
		}, Delay)

		self.tRunningTimers[nId] = ApolloTimer.Create(nDelay, false, "Trigger", tHandler)
	end

	function Encounter:RemoveDelay(strKey)
		if self.tRunningTimers[strKey] then
			self.tRunningTimers[strKey]:Stop()
			self.tRunningTimers[strKey] = nil
		end
	end
end

local metaInfSelfbuilder;
metaInfSelfbuilder = {__index = function(t,k)
	t[k] = setmetatable({}, metaInfSelfbuilder)
	return t[k]
end}
local idxCheck = {}
local idxNil = {}
function Encounter:Count(...)
	self.tVars.Counters = self.tVars.Counters or setmetatable({}, metaInfSelfbuilder)

	local tVal = self.tVars.Counters
	for i = 1, select('#', ...), 1 do
		local oIdx = select(i,...)
		if oIdx == nil then oIdx = idxNil end
		tVal = tVal[oIdx]
	end

	tVal[idxCheck] = (rawget(tVal, idxCheck) or 0)+1
	return rawget(tVal, idxCheck) --its easier this way... believe me.
end
function Encounter:GetCount(...)
	self.tVars.Counters = self.tVars.Counters or setmetatable({}, metaInfSelfbuilder)

	local tVal = self.tVars.Counters
	for i = 1, select('#', ...), 1 do
		local oIdx = select(i,...)
		if oIdx == nil then oIdx = idxNil end
		tVal = tVal[oIdx]
	end

	return rawget(tVal, idxCheck) or 0
end
function Encounter:SetCount(nVal, ...)
	self.tVars.Counters = self.tVars.Counters or setmetatable({}, metaInfSelfbuilder)

	local tVal = self.tVars.Counters
	for i = 1, select('#', ...), 1 do
		local oIdx = select(i,...)
		if oIdx == nil then oIdx = idxNil end
		tVal = tVal[oIdx]
	end

	tVal[idxCheck] = tonumber(nVal) --this throws no errors.
	return rawget(tVal, idxCheck)
end

function Encounter:IsRunning()
	return self.run
end

function Encounter:IsEnabled()
	return self.config.enable
end

function Encounter:InRange(tOrigin, nRange)
	return self.core:GetDistance(tOrigin) <= nRange
end

function Encounter:GetDistance(tOrigin)
	return self.core:GetDistance(tOrigin)
end

function Encounter:GetGroupUnits() --luacheck: no unused args
	if GroupLib.InGroup() then
		local tUnits = {}
		for nMember = 1, GroupLib.GetMemberCount(), 1 do
			local tUnit = GroupLib.GetUnitForGroupMember(nMember)
			if tUnit then
				table.insert(tUnits, tUnit)
			end
		end
		return tUnits
	else
		return {GameLib.GetPlayerUnit()}
	end
end

function Encounter:Init(parent)
	Apollo.LinkAddon(parent, self)

	self.core = parent
	self.L = parent:GetLocale(self.strEncounterKey, self.tLocalizations)

	self.strEncounterKey = nil
	self.tLocalizations = nil
end

Encounter.__index = Encounter
function Encounter.new(strEncounterKey, tMetaData, tLocalizations, fConfigs, fMethods)
	local o = setmetatable({
		strEncounterKey = strEncounterKey, 		--these two will be set to nil
		tLocalizations = tLocalizations, 		--after use! (garbage collection)

		run = false,
		runtime = {},
		instance = tMetaData.instance,
		groupName = tMetaData.groupName,
		displayName = tMetaData.displayName,
		tTrigger = tMetaData.tTrigger,

		tCallbacks = {},
		tEvents = {},
		tTrackedUnits = {},
		tCreatedUnits = {},
		tRunningTimers = {},
		tVars = {},
	}, Encounter)

  o.config = tAddon.ConfigEnv.Apply({enable = true}, fConfigs)
  tAddon.MethodEnv.Apply(o, fMethods)

	tAddon.Injector:Inject(strEncounterKey, o)
	return o
end